do
	local InitGlobalsOrigin = InitGlobals
	function InitGlobals()
		InitGlobalsOrigin()



----------------------------------------------- SquareArray
if not SquareArray then
local floor = math.floor
local ceil = math.ceil


local function round(x) return x > 0 and floor(x+0.5) or ceil(x-0.5) end

local function minmax(a, b)
	if a <= b then return a, b
	else return b, a end
end

local function is_x(self, x)
	return self.minX <= x and self.maxX >= x
end

local function is_y(self, y)
	return self.minY <= y and self.maxY >= y
end

local function norm_x(self, x)
	if x < self.minX or x > self.maxX then return nil end
	return x
end

local function norm_y(self, y)
	if y < self.minY or y > self.maxY then return nil end
	return y
end

local function norm_ix(self, ix)
	if ix < self.minIX or ix > self.maxIX then return nil end
	return ix
end

local function norm_iy(self, iy)
	if iy < self.minIY or iy > self.maxIY then return nil end
	return iy
end

local function norm_xx(self, minX, maxX)
	minX, maxX = minmax(minX, maxX)
	if maxX < self.minX or minX > self.maxX then return nil end
	if minX < self.minX then minX = self.minX end
	if maxX > self.maxX then maxX = self.maxX end
	return minX, maxX
end

local function norm_yy(self, minY, maxY)
	minY, maxY = minmax(minY, maxY)
	if maxY < self.minY or minY > self.maxY then return nil end
	if minY < self.minY then minY = self.minY end
	if maxY > self.maxY then maxY = self.maxY end
	return minY, maxY
end

local function norm_ixix(self, minIX, maxIX)
	minIX, maxIX = minmax(minIX, maxIX)
	if maxIX < self.minIX or minIX > self.maxIX then return nil end
	if minIX < self.minIX then minIX = self.minIX end
	if maxIX > self.maxIX then maxIX = self.maxIX end
	return minIX, maxIX
end

local function norm_iyiy(self, minIY, maxIY)
	minIY, maxIY = minmax(minIY, maxIY)
	if maxIY < self.minIY or minIY > self.maxIY then return nil end
	if minIY < self.minIY then minIY = self.minIY end
	if maxIY > self.maxIY then maxIY = self.maxIY end
	return minIY, maxIY
end

local index = {}

function index:norm_xy(x,y) return norm_x(self, x), norm_y(self, y) end

function index:is_xy(x, y) return norm_x(self, x) and norm_y(self, y) end

function index:rect_to_indexes(x0, y0, x1, y1)
	x0, x1 = norm_xx(self, x0, x1)
	y0, y1 = norm_yy(self, y0, y1)
	if not (x0 and y0) then return nil end
	x0, y0 = self:xy_to_ixiy(x0, y0)
	return x0, y0, self:xy_to_ixiy(x1, y1)
end

function index:xy_to_ixiy(x, y)
	x, y = self:norm_xy(x, y)
	if not (x and y) then return nil end
	local offsetX, offsetY, step = self.offsetX, self.offsetY, self.step
	local ix = round((x+offsetX)/step)
	local iy = round((y+offsetY)/step)
	return ix, iy
end

function index:get_xy(x, y)
	local ix, iy = self:xy_to_ixiy(x, y)
	return ix and iy and self.grid[iy][ix] or nil
end

function index:set_xy(x, y, val)
	local ix, iy = self:xy_to_ixiy(x, y)
	if ix and iy then self.grid[iy][ix] = val end
end

function index:get_ixiy(ix, iy)
	ix, iy = norm_ix(self, ix), norm_iy(self, iy)
	return ix and iy and self.grid[iy][ix] or nil
end

function index:set_ixiy(ix, iy, val)
	ix, iy = norm_ix(self, ix), norm_iy(self, iy)
	if ix and iy then self.grid[iy][ix] = val end
end

function index:id_xy(x, y)
	local ix, iy = self:xy_to_ixiy(x, y)
	return ix and iy and iy*self.maxIX+ix or nil
end

function index:by_all_square()
	local offsetX, offsetY, step = self.offsetX, self.offsetY, self.step
	local minIX, minIY, maxIX, maxIY = self.minIX, self.minIY, self.maxIX, self.maxIY
	-- print(offsetX, offsetY, step)
	-- print(minIX, minIY, maxIX, maxIY)
	local t_minIX = minIX - 1
	return function()
		t_minIX = t_minIX + 1
		if t_minIX > maxIX then
			t_minIX = minIX
			minIY = minIY + 1
		end
		if minIY <= maxIY then
			return t_minIX*step-offsetX, minIY*step-offsetY, t_minIX, minIY
		end
	end
end

function index:by_square(minX, minY, maxX, maxY)
	local offsetX, offsetY, step = self.offsetX, self.offsetY, self.step
	local minIX, maxIX = norm_xx(self, minX, maxX)
	local minIY, maxIY = norm_yy(self, minY, maxY)
	if not (minIX and minIY) then return nil end
	local t_minIX = minIX - 1
	return function()
		t_minIX = t_minIX + 1
		if t_minIX > maxIX then
			t_minIX = minIX
			minIY = minIY + 1
		end
		if minIY <= maxIY then
			return t_minIX*step-offsetX, minIY*step-offsetY, t_minIX, minIY
		end
	end
end

function index:by_isquare(minIX, minIY, maxIX, maxIY)
	local offsetX, offsetY, step = self.offsetX, self.offsetY, self.step
	minIX, maxIX = norm_ixix(self, minIX, maxIX)
	minIY, maxIY = norm_iyiy(self, minIY, maxIY)
	if not (minIX and minIY) then return nil end
	local t_minIX = minIX - 1
	return function()
		t_minIX = t_minIX + 1
		if t_minIX > maxIX then
			t_minIX = minIX
			minIY = minIY + 1
		end
		if minIY <= maxIY then
			return t_minIX*step-offsetX, minIY*step-offsetY, t_minIX, minIY
		end
	end
end

function index:by_circle(x0, y0, rds)
	rds = math.abs(rds)
	local offsetX, offsetY, step = self.offsetX, self.offsetY, self.step
	local minX, maxX = x0-rds, x0+rds
	local minY, maxY = y0-rds, y0+rds
	local t_minX = minX - step

	local opt_rds_2 = rds*rds
	local opt_y_2 = (minY-y0)*(minY-y0)
	return function()
		while true do
			t_minX = t_minX + step
			if t_minX > maxX then
				t_minX = minX
				while true do
					minY = minY + step
					if minY > maxY then return end
					if is_y(self, minY) then
						opt_y_2 = (minY-y0)*(minY-y0)
						break
					end
				end
			end
			if is_x(self, t_minX) and opt_y_2+(t_minX-x0)*(t_minX-x0) <= opt_rds_2 then
				return t_minX, minY, round((t_minX+offsetX)/step), round((minY+offsetY)/step)
			end
		end
	end
end


local mt = {__index = index}

function SquareArray(minX, minY, maxX, maxY, step)
	minX, maxX = minmax(minX, maxX)
	minY, maxY = minmax(minY, maxY)
	assert(minX < maxX and minY < maxY and step > 0)
	assert(step > 0 and maxX-minX > step and maxY-minY > step)
	local maxIX = floor((maxX-minX)/step)
	local maxIY = floor((maxY-minY)/step)
	local grid = {}
	local object = {
		minX=minX, maxX=maxIX*step, minY=minY, maxY=maxIY*step, step=step,
		offsetY = -minY, offsetX = -minX, players = {},
		minIX = 0, minIY = 0, maxIX = maxIX, maxIY = maxIY,
		grid = grid,
	}
	for i=0, object.maxIY do grid[i] = {} end
	return setmetatable(object, mt)
end
end

----------------------------------------------- TerrainSystem
if not Territory then
assert(SquareArray)

local table_empty = {}

local function minmax(a, b)
	if a <= b then return a, b
	else return b, a end
end

local function area_generator()
	local i = 0 return function() i = i + 1 return i end
end

-- ok
local function is_player(self, playerId)
	return self.players[playerId] and playerId
end
-- ok
local function cell_create(playerId, areaId)
	return {pid = playerId, aid = areaId}
end
-- ok
local function is_cell_player(cell, playerId)
	return cell and cell.pid==playerId
end
local function player_create(self, playerId)
	if self.players[playerId] then return end
	local pl = {
		pid = playerId,
		areas = {},
	}
	self.players[playerId] = pl
end

local function area_create(self, tbl)
	if not (tbl.x and tbl.y and tbl.radius
		and tbl.pid and is_player(self, tbl.pid)
		and tbl.count
		and tbl.minIX and tbl.maxIX and tbl.minIY and tbl.maxIY) then
		-- print('area_create error')
		return nil
	end
	local area = {
		aid = self.gen(),
		pid = tbl.pid,
		x = tbl.x,
		y = tbl.y,
		radius = tbl.radius,
		count = tbl.count,
		minIX = tbl.minIX,
		maxIX = tbl.maxIX,
		minIY = tbl.minIY,
		maxIY = tbl.maxIY,
	}
	self.players[area.pid].areas[area.aid] = area
	self.areas[area.aid] = area
	return area
end

local function area_add_cell(area, cell, ix, iy)
	area.count = area.count + 1
	if iy < area.minIY then area.minIY = iy end
	if iy > area.maxIY then area.maxIY = iy end
	if ix < area.minIX then area.minIX = ix end
	if ix > area.maxIX then area.maxIX = ix end
	-- print('cell', 'area '..area.aid, 'ix '..ix, 'iy '..iy)
end

local function area_delete(self, areaId)
	-- print('area_delete', 'start')
	local a = self.areas[areaId] -- area
	if not a or a.count > 0 then return true end
	-- for _, _, ix, iy in self.array:by_isquare(a.minIX, a.minIY, a.maxIX, a.maxIY) do
	-- 	local cell = array:get_ixiy(ix, iy)
	-- 	if cell and cell.aid == areaId then array:set_ixiy(ix, iy, nil)
	-- end
	self.players[a.pid].areas[areaId] = nil
	self.areas[areaId] = nil
	-- print('area_delete', 'end')
	return true
end
-- ok
local function update_area_size(self, areaId)
	local area, array = self.areas[areaId], self.array
	local count = 0
	local minIX, maxIX, minIY, maxIY = area.maxIX, area.minIX, area.maxIY, area.minIY
	for _, _, ix, iy in array:by_isquare(minIX, minIY, maxIX, maxIY) do
		local cell = array:get_ixiy(ix, iy)
		if cell and cell.aid == areaId then
			if iy < minIY then minIY = iy end
			if iy > maxIY then maxIY = iy end
			if ix < minIX then minIX = ix end
			if ix > maxIX then maxIX = ix end
			count = count + 1
		end
	end
	area.minIX, area.maxIX, area.minIY, area.maxIY = minIX, maxIX, minIY, maxIY
	area.count = count
end
-- ok
local function update_reduced_area(self, areaId)
	local area = self.areas[areaId]
	if not area then return end
	if area.count <= 0 then
		return area_delete(self, areaId)
	end
	update_area_size(self, areaId)
end
 -- ok
local function player_in_rect_one_more(self, playerId, ix0, iy0, ix1, iy1)
	local array = self.array
	for _, _, ix, iy in array:by_isquare(ix0,iy0,ix1,iy1) do
		local cell = array:get_ixiy(ix, iy) -- хотя бы 1 свой кусочек
		if is_cell_player(cell, playerId) then return true end
	end
	return false
end
 -- ok
local function empty_in_rect_one_more(self, ix0, iy0, ix1, iy1)
	local array = self.array
	for _, _, ix, iy in array:by_isquare(ix0,iy0,ix1,iy1) do
		local cell = array:get_ixiy(ix, iy)
		if not cell then return true end
	end
	return false
end
-- ok
local function rect_build(self, playerId, areaId, ix0, iy0, ix1, iy1)
	local array = self.array
	local area = self.areas[areaId]
	for _, _, ix, iy in array:by_isquare(ix0,iy0,ix1,iy1) do
		if not array:get_ixiy(ix, iy) then
			local cell = cell_create(playerId, areaId)
			array:set_ixiy(ix, iy, cell)
			area_add_cell(area, cell, ix, iy)
		end
	end
end
-- ok
local function rect_transfer(self, sellerId, buyerId, areaId, ix0, iy0, ix1, iy1)
	local array, touched_areas = self.array, {}
	local area = self.areas[areaId]
	for _, _, ix, iy in array:by_isquare(ix0,iy0,ix1,iy1) do
		local cell = array:get_ixiy(ix, iy)
		if cell then
			if cell.pid == sellerId then
				local aid = cell.aid
				area.count = area.count - 1
				touched_areas[aid] = true
				cell.pid, cell.aid = buyerId, areaId
				area_add_cell(area, cell, ix, iy)
			end
		else
			cell = cell_create(buyerId, areaId)
			array:set_ixiy(ix, iy, cell)
			area_add_cell(area, cell, ix, iy)
		end
	end
	for aid, _ in pairs(touched_areas) do
		update_reduced_area(self, aid)
	end
end
-- ok
local function near_area_to_rect(self, playerId, ix0, iy0, ix1, iy1)
	local array = self.array
	for _, _, ix, iy in array:by_isquare(ix0-1, iy0, ix0-1, iy1) or table_empty do
		local cell = array:get_ixiy(ix, iy)
		if is_cell_player(cell, playerId) then return cell.aid end
	end
	for _, _, ix, iy in array:by_isquare(ix0, iy1+1, ix1, iy1+1) or table_empty do
		local cell = array:get_ixiy(ix, iy)
		if is_cell_player(cell, playerId) then return cell.aid end
	end
	for _, _, ix, iy in array:by_isquare(ix1+1, iy0, ix1+1, iy1) or table_empty do
		local cell = array:get_ixiy(ix, iy)
		if is_cell_player(cell, playerId) then return cell.aid end
	end
	for _, _, ix, iy in array:by_isquare(ix0, iy0-1, ix1, iy0-1) or table_empty do
		local cell = array:get_ixiy(ix, iy)
		if is_cell_player(cell, playerId) then return cell.aid end
	end
	return nil
end
-- ok
local function is_rect_surrounded_player(self, playerId, ix0, iy0, ix1, iy1)
	local array = self.array
	for _, _, ix, iy in array:by_isquare(ix0-1, iy0, ix0-1, iy1) or table_empty do
		if not is_cell_player(array:get_ixiy(ix, iy), playerId) then return false end
	end
	for _, _, ix, iy in array:by_isquare(ix0, iy1+1, ix1, iy1+1) or table_empty do
		if not is_cell_player(array:get_ixiy(ix, iy), playerId) then return false end
	end
	for _, _, ix, iy in array:by_isquare(ix1+1, iy0, ix1+1, iy1) or table_empty do
		if not is_cell_player(array:get_ixiy(ix, iy), playerId) then return false end
	end
	for _, _, ix, iy in array:by_isquare(ix0, iy0-1, ix1, iy0-1) or table_empty do
		if not is_cell_player(array:get_ixiy(ix, iy), playerId) then return false end
	end
	return true
end
-- ok
local function max_area_player_rect(self, playerId, ix0, iy0, ix1, iy1)
	local array = self.array
	local max, tbl, areaId = 0, {}
	for _, _, ix, iy in array:by_isquare(ix0, iy0, ix1, iy1) do
		local cell = array:get_ixiy(ix, iy)
		if is_cell_player(cell, playerId) then
			local aid = cell.aid
			tbl[aid] = tbl[aid] and tbl[aid]+1 or 1
		end
	end
	for aid, num in pairs(tbl) do
		if max < num then max, areaId = num, aid end
	end
	return areaId
end

local function mark_border(self, pid, ix0, iy0, ix1, iy1)
	local array = self.array
	local start, first = nil, true -- first cell
	local count, mark, rect  = 0, {}, {}
	-- print('mark_border', ix0, iy0, ix1, iy1)
	for _, _, ix, iy in array:by_isquare(ix0-1, iy0, ix0-1, iy1) or table_empty do
		local cell = array:get_ixiy(ix0, iy) -- left
		mark[iy] = {}
		rect[iy] = {}
		if (not cell or is_cell_player(cell, pid))
			and not is_cell_player(array:get_ixiy(ix, iy), pid) then
			mark[iy][ix0] = true
			count = count + 1
			if first then start = {2, ix0, iy} first = nil end
		end
	end
	for _, _, ix, iy in array:by_isquare(ix0, iy1+1, ix1, iy1+1) or table_empty do
		local cell = array:get_ixiy(ix, iy1) -- top
		if (not cell or is_cell_player(cell, pid))
			and not is_cell_player(array:get_ixiy(ix, iy), pid) and not mark[iy1][ix] then
			mark[iy1][ix] = true
			count = count + 1
			if first then start = {3, ix, iy1} first = nil end
		end
	end
	for _, _, ix, iy in array:by_isquare(ix1+1, iy0, ix1+1, iy1) or table_empty do
		local cell = array:get_ixiy(ix1, iy) -- right
		if (not cell or is_cell_player(cell, pid))
			and not is_cell_player(array:get_ixiy(ix, iy), pid) and not mark[iy][ix1] then
			mark[iy][ix1] = true
			count = count + 1
			if first then start = {4, ix1, iy} first = nil end
		end
	end
	for _, _, ix, iy in array:by_isquare(ix0, iy0-1, ix1, iy0-1) or table_empty do
		local cell = array:get_ixiy(ix, iy0) -- bot
		if (not cell or is_cell_player(cell, pid))
			and not is_cell_player(array:get_ixiy(ix, iy), pid) and not mark[iy0][ix] then
			mark[iy0][ix] = true
			count = count + 1
			if first then start = {1, ix, iy0} first = nil end
		end
	end
	-- запонение прямоугольника новой области
	for _, _, ix, iy in array:by_isquare(ix0, iy0, ix1, iy1) do
		local cell = array:get_ixiy(ix, iy)
		if not cell then rect[iy][ix] = true end
	end
	-- local rect = {}
	-- for iy=iy0, iy1 do
	-- 	local trect = {}
	-- 	rect[iy] = trect
	-- 	for ix=ix0, ix1 do
	-- 		trect[ix] = true
	-- 		-- print('rect', ix, iy)
	-- 	end
	-- end
	if start then -- значит забагаюзали, убирается дырка
		return rect, mark, count, start[1], start[2], start[3]
	else
		return rect, mark, count
	end
end

local function stbl(tbl, x, y) return tbl[y] and tbl[y][x] end
local function bypass_border(self, pid, ix0, iy0, ix1, iy1)
	local array = self.array
	local rect, mark, count, state, x0, y0 = mark_border(self, pid, ix0, iy0, ix1, iy1)
	if not state then return true end
	local tx, ty = x0, y0
	local opt = 0
	-- print('x0, y0',x0, y0)
	repeat
		if stbl(mark, tx, ty) then
			mark[ty][tx] = nil
			count = count - 1
		end
		while true do
			opt = opt + 1
			if opt > 400 then print('bypass_border error opt') return end
			-- print('tx '..tx, 'tx '..ty, 'state '..state)
			if state == 1 then
				local cell = array:get_ixiy(tx-1, ty-1)
				if is_cell_player(cell, pid) or stbl(rect, tx-1, ty-1) then
					tx, ty, state = tx-1, ty-1, 4 break
				else
					cell = array:get_ixiy(tx-1, ty)
					if is_cell_player(cell, pid) or stbl(rect, tx-1, ty) then
						tx = tx-1 break
					else
						state = 2
					end
				end
			elseif state == 2 then
				local cell = array:get_ixiy(tx-1, ty+1)
				if is_cell_player(cell, pid) or stbl(rect, tx-1, ty+1) then
					tx, ty, state = tx-1, ty+1, 1 break
				else
					cell = array:get_ixiy(tx, ty+1)
					if is_cell_player(cell, pid) or stbl(rect, tx, ty+1) then
						ty = ty+1 break
					else
						state = 3
					end
				end
			elseif state == 3 then
				local cell = array:get_ixiy(tx+1, ty+1)
				if is_cell_player(cell, pid) or stbl(rect, tx+1, ty+1) then
					tx, ty, state = tx+1, ty+1, 2 break
				else
					cell = array:get_ixiy(tx+1, ty)
					if is_cell_player(cell, pid) or stbl(rect, tx+1, ty) then
						tx = tx+1 break
					else
						state = 4
					end
				end
			elseif state == 4 then
				local cell = array:get_ixiy(tx+1, ty-1)
				if is_cell_player(cell, pid) or stbl(rect, tx+1, ty-1) then
					tx, ty, state = tx+1, ty-1, 3 break
				else
					cell = array:get_ixiy(tx, ty-1)
					if is_cell_player(cell, pid) or stbl(rect, tx, ty-1) then
						ty = ty-1 break
					else
						state = 1
					end
				end
			end
		end
	until tx==x0 and ty==y0
	-- print('count', count)
	return count <= 0
end

-- basic
local function can_build(self, playerId, x0, y0, x1, y1)
	local minIX, minIY, maxIX, maxIY = self.array:rect_to_indexes(x0,y0,x1,y1)
	if not is_player(self, playerId)
		or not minIX
		or not empty_in_rect_one_more(self, minIX, minIY, maxIX, maxIY) then
		return false
	end
	local areaId = max_area_player_rect(self, playerId, minIX, minIY, maxIX, maxIY)
				or near_area_to_rect(self, playerId, minIX, minIY, maxIX, maxIY)
	-- print('can_build areaId', areaId)
	if not areaId then return false end
	if not bypass_border(self, playerId, minIX, minIY, maxIX, maxIY) then
		-- print('bypass_border error 2')
		return false
	end
	return true, areaId
end


-- basic
local function build_land(self, playerId, x0, y0, x1, y1)
	local minIX, minIY, maxIX, maxIY = self.array:rect_to_indexes(x0,y0,x1,y1)
	if not (minIX and is_player(self, playerId)) then return false end
	local can, areaId = self:can_build(playerId, x0, y0, x1, y1)
	if not can then return false end
	rect_build(self, playerId, areaId, minIX, minIY, maxIX, maxIY)
	return true
end


-- basic
local function sell_land(self, sellerId, buyerId, x0, y0, x1, y1)
	local minIX, minIY, maxIX, maxIY = self.array:rect_to_indexes(x0,y0,x1,y1)
	if not (minIX and is_player(self, sellerId) and is_player(self, buyerId))
		or is_rect_surrounded_player(self, sellerId, minIX, minIY, maxIX, maxIY)
		or not player_in_rect_one_more(self, sellerId, minIX, minIY, maxIX, maxIY) then
		return false
	end
	local areaId = near_area_to_rect(self, buyerId, minIX, minIY, maxIX, maxIY)
	if not areaId then
		areaId = area_create(self, {
			pid = buyerId,
			x = (x0+x1)*0.5,
			y = (y0+y1)*0.5,
			radius = 0,
			count = 0,
			minIX = minIX,
			maxIX = maxIX,
			minIY = minIY,
			maxIY = maxIY,
		}).aid
	end
	rect_transfer(self, sellerId, buyerId, areaId, minIX, minIY, maxIX, maxIY)
	return true
end

local function can_new_village(self, playerId, x0, y0, radius)
	-- ищем перекрывающиеся деревни
	for aId, tbl in pairs(self.areas) do
		local dist = (tbl.x-x0)*(tbl.x-x0)+(tbl.y-y0)*(tbl.y-y0)
		local rds = radius+tbl.radius
		if dist < rds*rds then
			-- print('dist error')
			return false
		end
	end
	-- смотрим не заняты ли клетки
	local array = self.array
	for _, _, ix, iy in array:by_circle(x0, y0, radius) do
		local cell = array:get_ixiy(ix, iy)
		if cell and cell.pid~=playerId then
			-- print('array error')
			return false
		end
	end
	return true
end

-- basic
local function new_village(self, playerId, x, y, rds)
	if not (playerId and x and y and rds)
		or not can_new_village(self, playerId, x, y, rds) then
		return false
	end
	player_create(self, playerId)
	local minIX, minIY, maxIX, maxIY = self.array:rect_to_indexes(x-rds,y-rds,x+rds,y+rds)
	local area = area_create(self, {
		x = x,
		y = y,
		radius = rds,
		pid = playerId,
		count = 0,
		minIX = minIX,
		maxIX = maxIX,
		minIY = minIY,
		maxIY = maxIY,
	})
	local array, areaId = self.array, area.aid
	for _, _, ix, iy in array:by_circle(x, y, rds) do
		-- print('v',ix,iy)
		local cell = cell_create(playerId, areaId)
		array:set_ixiy(ix, iy, cell)
		area_add_cell(area, cell, ix, iy)
	end
	if area.count <= 0 then
		print('error new_village')
		area_delete(self, area) --
		return false
	end
	-- print('areaId', areaId)
	return true
end

local function area_points(self, areaId)
	local array, a = self.array, self.areas[areaId]
	if not a then return nil end
	local couples = {}
	for x, y, ix, iy in array:by_isquare(a.minIX, a.minIY, a.maxIX, a.maxIY) do
		local cell = array:get_ixiy(ix, iy)
		if cell and cell.aid == areaId then
			table.insert(couples, x)
			table.insert(couples, y)
		end
	end
	return couples
end

local function for_sort(a,b) return a.id < b.id end
local function player_points(self, playerId)
	if not is_player(self, playerId) then return nil end
	local areas = self.players[playerId].areas
	local list = {}
	for areaId, area in pairs(areas) do -- desync
		table.insert(list, {id=area.aid, data = area_points(self, areaId)})
	end
	table.sort(list, for_sort)
	return list
end

local index = {
	can_build = can_build,
	build_land = build_land,
	sell_land = sell_land,
	new_village = new_village,
	player_points = player_points,
}

local mt = {__index = index}

function Territory(minX, minY, maxX, maxY, step)
	local array = SquareArray(minX, minY, maxX, maxY, step)
	local object = {
		step = step, offsetY = -minY, offsetX = -minX,
		array = array,
		areas = {}, players = {},
		gen = area_generator()
	}
	return setmetatable(object, mt)
end
end

----------------------------------------------- Пример использования Territory
do
local world_rect = GetWorldBounds()
-- получаем все значение карты
local minX, maxX, minY, maxY = GetRectMinX(world_rect), GetRectMaxX(world_rect), GetRectMinY(world_rect), GetRectMaxY(world_rect)
local step = 64 -- шаг сетки
local map = Territory(minX, minY, maxX, maxY, step)
local home_radius = 150 -- радиус деревни
local buy_radius = step/2 -- радиус покупки земли


-------------------------- пример - отображение в игре
local player_units = {[1]={},[2]={}} -- храним юнитов отображения для plyerId-1 и plyerId-2

-- обновляем ландшафт в точке (пример)
local function update_land_ex(pId, x, y)
	local red = pId==1 and 255 or 150
	local blue = pId==2 and 255 or 150
	local green = 150
	local alpha = 100
	local size = 64 -- размер юнита
	local scale = step/size
	-- 1747988528 это его UID
	local uni = CreateUnit(Player(pId), 1747988528, x, y, 0)
	SetUnitScale(uni, scale, scale, scale)
	SetUnitVertexColor(uni, red, blue, green, alpha)
	return uni
end

-- обновляем отображения для playerId
local function update_land(playerId)
	local tbl = {}
	local pl_areas = map:player_points(playerId)
	for _, area in ipairs(pl_areas) do -- идем по всем областям (area) игрока (playerId)
		area = area.data
		for i=1, #area, 2 do
			-- area[i], area[i+1] это x и y
			tbl[65536*area[i+1]+area[i]] = {area[i], area[i+1]}
		end
	end
	local dict = player_units[playerId]
	--// десинх
	for k, uni in pairs(dict) do if not tbl[k] then RemoveUnit(uni) dict[k] = nil end end -- тут удаляю юнитов которые больше не нужно отображать (квадратики)
	for k, couple in pairs(tbl) do if not dict[k] then dict[k] = update_land_ex(playerId, couple[1], couple[2]) end end -- тут дорисовываю новых юнитов
	--//
end
--------------------------

-- Триггер который регистрирует нажатия на способности
-- 1 и 2 это тестовые plyerId
local trig = CreateTrigger()
TriggerRegisterAnyUnitEventBJ( trig, EVENT_PLAYER_UNIT_SPELL_EFFECT )
TriggerAddAction( trig, function()
	local aid, tx, ty = GetSpellAbilityId(), GetSpellTargetX(), GetSpellTargetY()
	local basic = 1093677104 -- это просто id способности
	local result
	print('tx,ty',tx,ty)
	if aid==basic then -- create 1
		print('create 1')
		result = map:new_village(1, tx, ty, home_radius, step)
		if result then update_land(1) end
		print(result)
	elseif aid==basic+1 then -- create 2
		print('create 2')
		result = map:new_village(2, tx, ty, home_radius, step)
		if result then update_land(2) end
		print(result)
	elseif aid==basic+2 then -- buy 1
		print('buy 1')
		result = map:build_land(1, tx-buy_radius,ty-buy_radius,tx+buy_radius,ty+buy_radius)
		if result then update_land(1) end
		print(result)
	elseif aid==basic+3 then -- buy 2
		print('buy 2')
		result = map:build_land(2, tx-buy_radius,ty-buy_radius,tx+buy_radius,ty+buy_radius)
		if result then update_land(2) end
		print(result)
	elseif aid==basic+4 then -- send 1 -> 2
		print('send 1 -> 2')
		result = map:sell_land(1,2, tx-buy_radius,ty-buy_radius,tx+buy_radius,ty+buy_radius)
		if result then update_land(1) update_land(2) end
		print(result)
	elseif aid==basic+5 then -- send 2 -> 1
		print('send 2 -> 1')
		result = map:sell_land(2,1, tx-buy_radius,ty-buy_radius,tx+buy_radius,ty+buy_radius)
		if result then update_land(1) update_land(2) end
		print(result)
	end
end)

end
-----------------------------------------------

	end
end